/* Fichier "camera.c"
 * Routines de gestion de la camera (placement, deplacement...)
 * 
 * On aurait pu s'en passer, mais : 
 * 1) ca rend les choses plus intuitives de placer sa 
 *    camra  telle position et de regarder dans telle direction,
 *    plutot que d'utiliser rotations et translations
 * 2) ca pourra devenir utile si l'on veut se deplacer dans 
 *    le terrain par la suite 
 */

#include "projet.h"

/* #define CAMERA_DEBUG */

static camera_t camera;

/* ======================================================
 * cam_init
 *
 * Description: initialisation de la camera, placement 
 *   l'origine
 * =====================================================
 */
void
cam_init ()
{
  /* initialisation basique des 3 vecteurs de la camera */
  set_vector (&camera.position, 0.0, 0.0, 0.0);
  set_vector (&camera.view, 0.0, 1.0, 0.5);
  set_vector (&camera.upVector, 0.0, 0.0, 1.0);
  
  camera.control_mode = CAMERA_CONTROL_TERRAIN;

#ifdef CAMERA_DEBUG
  g_print ("camera: init\n");
#endif
}

/* ======================================================
 * cam_set_default_view
 *
 * Description: positionne la camra au point x=0, y=60, z=150
 * =====================================================
 */
void
cam_set_default_view ()
{
  vector3d_t pos, view, upvect;
  set_vector (&pos, 0.0f, 60.0f, 150.0f);
  set_vector (&view, 0.0f, 0.0f, 0.0f);
  upvect = cam_get_upvector (view, pos);

  cam_set_posv (pos, view, upvect);
}

/* ======================================================
 * cam_switch_control_mode
 *
 * Description: 
 * =====================================================
 */
void
cam_switch_control_mode ()
{
  if (camera.control_mode == CAMERA_CONTROL_CAMERA)
    camera.control_mode = CAMERA_CONTROL_TERRAIN;
  else camera.control_mode = CAMERA_CONTROL_CAMERA;
  
#ifdef CAMERA_DEBUG
  g_print ("camera: control mode = %d\n", camera.control_mode);
#endif
}

void
cam_set_control_mode (int newmode)
{
  if (newmode == CAMERA_CONTROL_CAMERA || newmode == CAMERA_CONTROL_TERRAIN)
    camera.control_mode = newmode;
}

int
cam_get_control_mode ()
{
  return camera.control_mode;
}

/* ======================================================
 * cam_get_upvector
 *
 * Description: retourne le vecteur pointant vers le haut de la
 *  camra,  partir des de la position et du point regard
 * Parametres: point regard et position
 * =====================================================
 */
vector3d_t
cam_get_upvector (vector3d_t view, vector3d_t pos)
{
  vector3d_t cross, vpos;

  set_vector (&vpos, view.coord[0] - pos.coord[0],
	      view.coord[1] - pos.coord[1], view.coord[2] - pos.coord[2]);

  cross.coord[0] = -vpos.coord[0] * vpos.coord[1];
  cross.coord[1] =
    vpos.coord[0] * vpos.coord[0] + vpos.coord[2] * vpos.coord[2];
  cross.coord[2] = -vpos.coord[1] * vpos.coord[2];

  normalize (&cross);
#ifdef CAMERA_DEBUG
  g_print ("upvector: (x=%f y=%f z=%f)\n", cross.coord[0],
	   cross.coord[1], cross.coord[2]);
#endif
  return cross;
}


/* ======================================================
 * cam_set_posf
 *
 * Description: permet de placer la camera et sa direction 
 *  de vue en utilisant des floats 
 * Parametres: coord. x, y et z de la position, vue et vect. normal 
 * =====================================================
 */
void
cam_set_posf (float positionX, float positionY, float positionZ,
	      float viewX, float viewY, float viewZ,
	      float upVectorX, float upVectorY, float upVectorZ)
{
  set_vector (&camera.position, positionX, positionY, positionZ);
  set_vector (&camera.view, viewX, viewY, viewZ);
  set_vector (&camera.upVector, upVectorX, upVectorY, upVectorZ);

#ifdef CAMERA_DEBUG
  g_print ("camera: placed at x=%f, y=%f, z=%f\n", positionX, positionY,
	   positionZ);
#endif
}

/* ======================================================
 * cam_set_posv
 *
 * Description: permet de placer la camera et sa direction 
 *  de vue en utilisant directement des vecteurs 
 *  Parametres: vecteurs de la position, vue et normal 
 * =====================================================
 */
void
cam_set_posv (vector3d_t pos, vector3d_t view, vector3d_t upvect)
{
  camera.position = pos;
  camera.view = view;
  camera.upVector = upvect;
#ifdef CAMERA_DEBUG
  g_print ("camera: placed at x=%f, y=%f, z=%f\n", camera.position.coord[0],
	   camera.position.coord[1], camera.position.coord[2]);
#endif
}

/* ======================================================
 * cam_look
 *
 * Description: met  jour les infos issues de la camera pour le rendu 
 * =====================================================
 */
void
cam_look ()
{
  gluLookAt (camera.position.coord[0], camera.position.coord[1],
	     camera.position.coord[2],
	     camera.view.coord[0], camera.view.coord[1],
	     camera.view.coord[2],
	     camera.upVector.coord[0], camera.upVector.coord[1],
	     camera.upVector.coord[2]);
}

/* ======================================================
 * cam_move_cam_forwbackw
 *
 * Description: avance ou recule la camera  la vitesse 'speed' 
 * =====================================================
 */
void
cam_move_cam_forwbackw (float speed)
{
  vector3d_t newpos, posvect;

  set_vector (&posvect, camera.view.coord[0] - camera.position.coord[0],
	      camera.view.coord[1] - camera.position.coord[1],
	      camera.view.coord[2] - camera.position.coord[2]);
  normalize (&posvect);

  newpos = camera.position;
  newpos.coord[0] += speed * (posvect.coord[0]);
  newpos.coord[1] += speed * (posvect.coord[1]);
  newpos.coord[2] += speed * (posvect.coord[2]);

  cam_set_posv (newpos, camera.view, camera.upVector);
}

/* ======================================================
 * cam_move_cam_updown
 *
 * Description: monte ou descend la camera  la vitesse 'speed' 
 * =====================================================
 */
void
cam_move_cam_updown (float speed)
{
  vector3d_t upvectevect, posvect, newpos, newview;

  set_vector (&posvect, camera.view.coord[0] - camera.position.coord[0],
	      camera.view.coord[1] - camera.position.coord[1],
	      camera.view.coord[2] - camera.position.coord[2]);

  upvectevect = camera.upVector;
  normalize (&upvectevect);

  newpos = camera.position;
  newpos.coord[0] += speed * (upvectevect.coord[0]);
  newpos.coord[1] += speed * (upvectevect.coord[1]);
  newpos.coord[2] += speed * (upvectevect.coord[2]);

  newview = camera.view;
  newview.coord[0] += speed * (upvectevect.coord[0]);
  newview.coord[1] += speed * (upvectevect.coord[1]);
  newview.coord[2] += speed * (upvectevect.coord[2]);

  cam_set_posv (newpos, newview, camera.upVector);
}

/* ======================================================
 * cam_strafe
 *
 * Description: deplacement latral gauche/droite  la vitesse 'speed' 
 * =====================================================
 */
void
cam_strafe_cam_leftright (float speed)
{
  vector3d_t strafevect, posvect, newpos, newview;

  set_vector (&posvect, camera.view.coord[0] - camera.position.coord[0],
	      camera.view.coord[1] - camera.position.coord[1],
	      camera.view.coord[2] - camera.position.coord[2]);

  strafevect = cross_product_vect (posvect, camera.upVector);

#ifdef CAMERA_DEBUG
  g_print ("view=(%f,%f,%f), up=(%f,%f,%f), strafe=(%f,%f,%f)\n",
	   posvect.coord[0], posvect.coord[1], posvect.coord[2],
	   camera.upVector.coord[0], camera.upVector.coord[1],
	   camera.upVector.coord[2], strafevect.coord[0], strafevect.coord[1],
	   strafevect.coord[2]);
#endif

  normalize (&strafevect);

  newpos = camera.position;
  newpos.coord[0] += speed * (strafevect.coord[0]);
  newpos.coord[1] += speed * (strafevect.coord[1]);
  newpos.coord[2] += speed * (strafevect.coord[2]);

  newview = camera.view;
  newview.coord[0] += speed * (strafevect.coord[0]);
  newview.coord[1] += speed * (strafevect.coord[1]);
  newview.coord[2] += speed * (strafevect.coord[2]);

  cam_set_posv (newpos, newview, camera.upVector);
}

void
cam_rotate_terrain_leftright (float speed)
{
  vector3d_t newpos, newup, newposvect, posvect;
  float cosTheta = cos (speed);
  float sinTheta = sin (speed);

  set_vector (&posvect, camera.view.coord[0] - camera.position.coord[0],
	      camera.view.coord[1] - camera.position.coord[1],
	      camera.view.coord[2] - camera.position.coord[2]);

  newposvect.coord[0] = posvect.coord[0] * cosTheta + posvect.coord[2] * sinTheta;
  newposvect.coord[1] = posvect.coord[1];
  newposvect.coord[2] = -posvect.coord[0] * sinTheta
    + posvect.coord[2] * cosTheta;

  newpos.coord[0] = camera.view.coord[0]-newposvect.coord[0];
  newpos.coord[1] = camera.view.coord[1]-newposvect.coord[1];
  newpos.coord[2] = camera.view.coord[2]-newposvect.coord[2];
  
  newup = cam_get_upvector (camera.view, newpos);

  cam_set_posv (newpos, camera.view, newup);
}

void
cam_rotate_cam_leftright (float speed)
{
  vector3d_t newview, newup, upvect, newposvect, posvect;
  vector3d_t strafevect;

  float posv_length, strafev_length;
  float cosTheta = cos (speed);
  float sinTheta = sin (speed);

  set_vector (&posvect, camera.view.coord[0] - camera.position.coord[0],
	      camera.view.coord[1] - camera.position.coord[1],
	      camera.view.coord[2] - camera.position.coord[2]);
    
  upvect = cam_get_upvector (camera.view, camera.position);
  strafevect = cross_product_vect (posvect, camera.upVector);
  
  posv_length = get_vect_length(posvect);
  strafev_length = get_vect_length(strafevect);
  
  newposvect.coord[0] = posvect.coord[0] * cosTheta - (posv_length/strafev_length) *strafevect.coord[0] * sinTheta;
  newposvect.coord[1] = posvect.coord[1] * cosTheta - (posv_length/strafev_length) *strafevect.coord[1] * sinTheta;
  newposvect.coord[2] = posvect.coord[2] * cosTheta - (posv_length/strafev_length) *strafevect.coord[2] * sinTheta;

  newview.coord[0] = camera.position.coord[0]+newposvect.coord[0];
  newview.coord[1] = camera.position.coord[1]+newposvect.coord[1];
  newview.coord[2] = camera.position.coord[2]+newposvect.coord[2];
  
  newup = cam_get_upvector (newview, camera.position);

#ifdef CAMERA_DEBUG
  g_print("posvect=(%f,%f,%f), newposvect=(%f,%f,%f)\n",
    posvect.coord[0],posvect.coord[1],posvect.coord[2],
    newposvect.coord[0],newposvect.coord[1],newposvect.coord[2]);
#endif

  cam_set_posv (camera.position, newview, newup);
}

void
cam_rotate_cam_updown (float speed)
{
  vector3d_t newview, newup, upvect, newposvect, posvect;

  float posv_length, upv_length;
  float cosTheta = cos (speed);
  float sinTheta = sin (speed);

  set_vector (&posvect, camera.view.coord[0] - camera.position.coord[0],
	      camera.view.coord[1] - camera.position.coord[1],
	      camera.view.coord[2] - camera.position.coord[2]);
    
  upvect = cam_get_upvector (camera.view, camera.position);
  
  posv_length = get_vect_length(posvect);
  upv_length = get_vect_length(upvect);
  
  newposvect.coord[0] = posvect.coord[0] * cosTheta - (posv_length/upv_length) *upvect.coord[0] * sinTheta;
  newposvect.coord[1] = posvect.coord[1] * cosTheta - (posv_length/upv_length) *upvect.coord[1] * sinTheta;
  newposvect.coord[2] = posvect.coord[2] * cosTheta - (posv_length/upv_length) *upvect.coord[2] * sinTheta;

  newview.coord[0] = camera.position.coord[0]+newposvect.coord[0];
  newview.coord[1] = camera.position.coord[1]+newposvect.coord[1];
  newview.coord[2] = camera.position.coord[2]+newposvect.coord[2];
  
  newup = cam_get_upvector (newview, camera.position);

#ifdef CAMERA_DEBUG
  g_print("posvect=(%f,%f,%f), newposvect=(%f,%f,%f)\n",
    posvect.coord[0],posvect.coord[1],posvect.coord[2],
    newposvect.coord[0],newposvect.coord[1],newposvect.coord[2]);
#endif

  cam_set_posv (camera.position, newview, newup);
}
