/* Fichier "filters.c"
 * 
 * contient tous les filtres applicables
 *
 * En gnral, la fonction prend deux parametres :
 *   un pointeur sur le terrain  filtrer, et un parametre decrivant un effet
 *   (exemple: densit de cratres).
 *
 * Cette regle ne s'applique pas aux filtres  usage exclusivement interne
 *  (normalisation des hauteurs, par exemple).
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "terrain.h"
#include "common.h"

/* ======================================================
 * TER_FIL_FIR
 *
 * Description: algo FIR de Robert Krten (lissage)
 * Parametres: k : entre 0 (aucun lissage) et 1 (lissage maximal : plan)
 * =====================================================
 */
void
TER_FIL_FIR (TER_terrain_t * terrain, float k)
{
  int i = 0, j = 0;
  float z = 0.0f, old_z = 0.0f;

  /* lignes */
  old_z = TER_get_heightmap_value (terrain, i, j);
  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      z = k * old_z + (1 - k) * z;
      TER_set_heightmap_value (terrain, i, j, z);
      old_z = z;
    }
  }

  i = 0;
  j = 0;
  /* colonnes */
  old_z = TER_get_heightmap_value (terrain, i, j);
  for (j = 0; j < terrain->size; j++) {
    for (i = 0; i < terrain->size; i++) {
      z = TER_get_heightmap_value (terrain, i, j);
      z = k * old_z + (1 - k) * z;
      TER_set_heightmap_value (terrain, i, j, z);
      old_z = z;
    }
  }

  TER_update_maxheight_value (terrain);

  printf ("FILTER: <FIR> with param %f\n", k);
}

/* ======================================================
 * TER_FIL_normalize
 *
 * Description: met toutes les altitudes entre 0 et 1
 * =====================================================
 */
void
TER_FIL_normalize (TER_terrain_t * terrain)
{
  int i = 0, j = 0;
  float z = 0.0f;

  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      if (z < 0)
	z = 0;
      else
	z = z / terrain->max_height;
      TER_set_heightmap_value (terrain, i, j, z);
    }
  }
}

/* ======================================================
 * TER_FIL_denormalize
 *
 * Description: multiplie toutes les altitudes (supposes etre normalisees)
 *              par terrain->max_height
 *
 * =====================================================
 */
void
TER_FIL_denormalize (TER_terrain_t * terrain)
{
  int i = 0, j = 0;
  float z = 0.0f;

  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      z = z * terrain->max_height;
      TER_set_heightmap_value (terrain, i, j, z);
    }
  }
}

/* ======================================================
 * TER_FIL_power
 *
 * Description: eleve les altitudes  la puissance p
 *
 * =====================================================
 */
void
TER_FIL_power (TER_terrain_t * terrain, float p)
{
  int i = 0, j = 0;
  float z = 0.0f;

  TER_FIL_normalize (terrain);

  /* la valeur de maxheight ne change pas (=1) */

  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      z = (float) pow ((double) z, (double) p);
      TER_set_heightmap_value (terrain, i, j, z);
    }
  }

  TER_FIL_denormalize (terrain);

  printf ("FILTER: <power> with param %f\n", p);
}

/* ======================================================
 * TER_FIL_allpositive
 *
 * Description: supprime toutes les altitudes negatives
 *             (si z <0 alors z = 0) 
 *
 * =====================================================
 */
void
TER_FIL_allpositive (TER_terrain_t * terrain)
{
  int i = 0, j = 0;
  float z = 0.0f;

  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      if (z < 0) {
	z = 0;
	TER_set_heightmap_value (terrain, i, j, z);
      }
    }
  }
  
  TER_update_maxheight_value (terrain);
}

/* ======================================================
 * TER_FIL_translate_down
 *
 * Description: descend le terrain de la plus petite altitude 
 *   pour le centrer sur z=0 
 *
 * =====================================================
 */
void
TER_FIL_translate_down (TER_terrain_t * terrain)
{
  int i = 0, j = 0;
  float z = 0.0f;
  float zmin=terrain->max_height;

  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      if (z < zmin) {
	  zmin=z;
      }
    }
  }
  
  if (zmin > 0) {
  
  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      TER_set_heightmap_value (terrain, i, j, z-zmin);
    }
  }
  
  TER_update_maxheight_value (terrain);
}
}

/* ======================================================
 * TER_FIL_multiply
 *
 * Description: multiplie toutes les altitudes par m
 *
 * =====================================================
 */
void
TER_FIL_multiply (TER_terrain_t * terrain, float m)
{
  int i = 0, j = 0;
  float z = 0.0f;

  for (i = 0; i < terrain->size; i++) {
    for (j = 0; j < terrain->size; j++) {
      z = TER_get_heightmap_value (terrain, i, j);
      if (z < 0)
	z = 0;
      else
	z *= m;
      TER_set_heightmap_value (terrain, i, j, z);
    }
  }

  TER_update_maxheight_value (terrain);
  printf ("FILTER: <multiply> with param %f\n", m);
}

/* ======================================================
 * TER_FIL_craters
 *
 * Description: cree des crateres dans tout le terrain
 * Parametres: density: entre 1 (peu de crateres) 
 *     et 100 (recouvrent presque tout le terrain)
 *
 * =====================================================
 */
void
TER_FIL_craters (TER_terrain_t * terrain, float density)
{
  int x1, y1, x2, y2;
  float z = 0, newz = 0;
  float radius = 1.0f;
  int iter = 0;
  int num_crat = 0;

#define TERR_CRAT_RADIUS_MIN 3.0f
#define TERR_CRAT_RADIUS_MAX 10.0f
  
  srand (time (NULL));
  
  /* formule prenant en compte la taille du terrain, la taille maximale des 
     crateres et le paremetre density entre 0 et 100, et surtout,
     donnant des valeurs interessantes */
  num_crat = (int)((density/100) *
  ((terrain->size*terrain->size) / (0.8*TERR_CRAT_RADIUS_MAX*TERR_CRAT_RADIUS_MAX)));
  
  for (iter = 0; iter < num_crat; iter++) {

    x1 = random_int (terrain->size - 1);
    y1 = random_int (terrain->size - 1);
    radius = random_float (TERR_CRAT_RADIUS_MIN, TERR_CRAT_RADIUS_MAX);

    for (x2 = 0; x2 < terrain->size; x2++) {
      for (y2 = 0; y2 < terrain->size; y2++) {
	z = radius * radius - ((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));

	if (z < 0)
	  z = 0;
	else
	  z = sqrt (z);

	newz = TER_get_heightmap_value (terrain, x2, y2) - z;
	TER_set_heightmap_value (terrain, x2, y2, newz);
      }
    }
  }

  TER_update_maxheight_value (terrain);

  printf ("FILTER: <craters> with param %f\n", density);
}

/* ======================================================
 * TER_FIL_river
 *
 * Description: cree une riviere de largeur moyenne L
 *  ATTENTION: ce filtre n'est pas termin, il y a juste le dbut
 *
 * =====================================================
 */
void
TER_FIL_river (TER_terrain_t * terrain, int L)
{
  int cote1, cote2;
  int i_c11 = 0, i_c12 = 0, j_c11 = 0, j_c12 = 0,
    i_c21 = 0, i_c22 = 0, j_c21 = 0, j_c22 = 0;

  cote1 = random_int (4);
  do {
    cote2 = random_int (4);
  }
  while (cote2 == cote1);


  switch (cote1) {
  case 0:
    i_c11 = random_int (terrain->size - L);
    j_c11 = 0;
    i_c12 = L + i_c11;
    j_c12 = 0;
    break;
  case 1:
    i_c11 = terrain->size;
    j_c11 = random_int (terrain->size - L);
    i_c12 = terrain->size;
    j_c12 = L + j_c11;
    break;
  case 2:
    i_c11 = random_int (terrain->size - L);
    j_c11 = terrain->size;
    i_c12 = L + i_c11;
    j_c12 = terrain->size;
    break;
  case 3:
    i_c11 = 0;
    j_c11 = random_int (terrain->size - L);
    i_c12 = 0;
    j_c12 = L + j_c11;
    break;
  }

  switch (cote2) {
  case 0:
    i_c21 = random_int (terrain->size - L);
    j_c21 = 0;
    i_c22 = L + i_c21;
    j_c22 = 0;
    break;
  case 1:
    i_c21 = terrain->size;
    j_c21 = random_int (terrain->size - L);
    i_c22 = terrain->size;
    j_c22 = L + j_c21;
    break;
  case 2:
    i_c21 = random_int (terrain->size - L);
    j_c21 = terrain->size;
    i_c22 = L + i_c21;
    j_c22 = terrain->size;
    break;
  case 3:
    i_c21 = 0;
    j_c21 = random_int (terrain->size - L);
    i_c22 = 0;
    j_c22 = L + j_c21;
    break;
  }

  printf
    ("Cote_1: (i=%d,j=%d) (i=%d,j=%d) / Cote_2: (i=%d,j=%d) (i=%d,j=%d)\n",
     i_c11, j_c11, i_c12, j_c12, i_c21, j_c21, i_c22, j_c22);

  printf ("FILTER: <river> with param %d\n", L);
}
